; ++++++++++ Temperuture and Humidity sensor DHT22, Arduino NANO ++++++++++++

; PD0           - RXD UART 57600
; PD1           - TXD UART 57600
; PD4||PD5||PD6 - +5V for DHT22
; PD7           - BUS 1Wire DHT22

; PB5           - LED

; Transfer protocol - 6 byte every 30 seconds:
; NN,nn,HH,hh,TT,tt
; NN - MSB Transfer counter (0000, 0001, ... FFFF, 0000...)
; nn - LSB Transfer counter ///////////////////////////////
; HH - MSB Humidity (FFFF - Bus error FEFE -  CRC error)
; hh - LSB Humidity ////////////////////////////////////
; TT - MSB Signed Temperature (FFFF - Bus error FEFE -  CRC error)
; tt - LSB Temperature        ////////////////////////////////////

; R14-MSB, R15-LSB - Transfer USART counter
; R16 - MSB_RH, R17 - LSB_RH (FF,FF - Bus error FE,FE -  CRC error)
; R18 - MSB_T, R19 - LSB_T (FF,FF - Bus error FE,FE -  CRC error)
; R20 - CRC, Counter
; R21, R22, R23 - Counters
; R24 - Error counter

.include "m328def.inc"

.equ  DHT22       = 7;              ; PD7
.equ  DDR_DHT22   = DDRD;
.equ  PortDHT22   = PORTD;
.equ  PinDHT22    = PIND;

.equ  LED         = 5               ; PB5
.equ  PortLED     = PORTB;

.equ  MaxErr      = 3               ; max error reading DHT22

         .CSEG
         .ORG 0x0000

; ================ Start program =================
.include "Init328.asm"                  ; init SFR

          ldi    R23,55
WAIT1S1:  rcall  delay18ms
          dec    R23
          brne   WAIT1S1              ; WAIT ~1 sec
          wdr
          ldi    R23,55
WAIT1S2:  rcall  delay18ms
          dec    R23
          brne   WAIT1S2              ; WAIT ~1 sec
          wdr

          in     R23,PortDHT22
          ori    R23,0x70            ; 0111.0000
          out    PortDHT22,R23       ; ON +5V for DHT22

          ldi    R23,55
WAIT1S3:  rcall  delay18ms
          dec    R23
          brne   WAIT1S3             ; WAIT ~1 sec
          wdr
          ldi    R23,55
WAIT1S4:  rcall  delay18ms
          dec    R23
          brne   WAIT1S4             ; WAIT ~1 sec
          wdr
          ldi    R23,55
WAIT1S5:  rcall  delay18ms
          dec    R23
          brne   WAIT1S5             ; WAIT ~1 sec
          wdr
          clr    R24                 ; Error counter
          clr    R14                 ; Transfer USART counter
          clr    R15                 ; //

; PortOUT - Bus = 0 for 18 ms
RDDHT:    cbi    PortLED,LED         ; LED off
          sbis   PinDHT22,DHT22      ; BUS = 1, OK
          rjmp   ErrBus              ; Bus Error
          sbi    DDR_DHT22,DHT22     ; PORT = OUT
          cbi    PortDHT22,DHT22     ; PORT = 0
          rcall  delay18ms           ; Delay 18 ms
          cbi    DDR_DHT22,DHT22     ; PORT = IN

; PortIN - Wait for BUS=0 20-40 mks (Time Out ~150 mks)
          ldi    R21,115
WAIT1:    rcall  delay1mks           ; Delay 1 mks
          dec    R21
          breq   ErrBus              ; Time Out, error
          sbic   PinDHT22,DHT22      ; BUS = 0, OK
          rjmp   WAIT1               ; BUS = 1, wait

; PortIN - Wait for BUS=1 80 mks (Time Out ~200 mks)
          ldi    R21,155
WAIT2:    rcall  delay1mks           ; Delay 1 mks
          dec    R21
          breq   ErrBus              ; Time Out, error
          sbis   PinDHT22,DHT22      ; BUS = 1, OK
          rjmp   WAIT2               ; BUS = 0, wait

; PortIN - Wait for BUS=0 80 mks (Time Out ~200 mks)
          ldi    R21,155
WAIT3:    rcall  delay1mks           ; Delay 1 mks
          dec    R21
          breq   ErrBus              ; Time Out, error
          sbic   PinDHT22,DHT22      ; BUS = 0, OK
          rjmp   WAIT3               ; BUS = 1, wait
          rjmp   STARTTRN

; Bus error
ErrBus:   ldi    R16, 0xFF
          ldi    R17, 0xFF
          ldi    R18, 0xFF
          ldi    R19, 0xFF
          inc    R24                 ; error counter
          rjmp   TXCOM

; Start transfer
STARTTRN: ldi    R22,40              ; bit counter

; PortIN - Wait for BUS=1 50 mks (Time Out ~130 mks)
WAIT6:    ldi    R21,100
WAIT4:    rcall  delay1mks           ; Delay 1 mks
          dec    R21
          breq   ErrBus              ; Time Out, error
          sbis   PinDHT22,DHT22      ; BUS = 1, OK
          rjmp   WAIT4               ; BUS = 0, wait

; PortIN - Wait for BUS=0 - Receive bit 28 mcs =0, 70 mks =1 (Time Out ~120 mks)
          clr    R21
WAIT5:    rcall  delay1mks           ; Delay 1 mks
          inc    R21
          cpi    R21,90
          brcc   ErrBus              ; Time Out, R21>=90, error
          sbic   PinDHT22,DHT22      ; BUS = 0, receive bit
          rjmp   WAIT5               ; BUS = 1, wait
          cpi    R21,12
          brcs   ErrBus              ; R21<15, error
          cpi    R21,35              ; C=1 if R21<35, C=0 if R21>=50
          rol    R20                 ; save bit
          rol    R19                 ; //
          rol    R18                 ; //PIND 4 9
          rol    R17                 ; //
          rol    R16                 ; //
          dec    R22                 ; Bit counter
          brne   WAIT6               ; Receive next bit
          com    R16                 ; (R16...R20) = ~(R16...R20)
          com    R17                 ; //
          com    R18                 ; //
          com    R19                 ; //
          com    R20                 ; //

; End transfer from DHT22, Check CRC
          sub    R20,R19
          sub    R20,R18
          sub    R20,R17
          sub    R20,R16
          brne   ErrCRC
          clr    R24               ; Error counter
          rjmp   TXCOM             ; CRC OK

; CRC Error
ErrCRC:   ldi    R16, 0xFE
          ldi    R17, 0xFE
          ldi    R18, 0xFE
          ldi    R19, 0xFE
          inc    R24               ; error counter

; R14-MSB, R15-LSB - Transfer counter
; R16 - MSB_RH, R17 - LSB_RH
; R18 - MSB_T, R19 - LSB_T
; Send data to COM
TXCOM:    mov    R21,R14
          rcall  TXUART
          mov    R21,R15
          rcall  TXUART
          mov    R21,R16
          rcall  TXUART
          mov    R21,R17
          rcall  TXUART
          mov    R21,R18
          rcall  TXUART
          mov    R21,R19
          rcall  TXUART

          ldi    R21,0x01
          add    R15,R21           ; INC Transfer counter - LSB
          ldi    R21,0x00
          adc    R14,R21           ; MSB

          cpi    R24,MaxErr
          brcc   ERRORDHT          ; R24>=MaxErr

; All OK, delay 30 sec and new read from DHT22
DEL30:    ldi    R20,30
DEL304:   ldi    R23,20
DEL303:   ldi    R22,200
DEL302:   ldi    R21,210
DEL301:   rcall  delay1mks
          dec    R21
          brne   DEL301            ; ~250 mks
          dec    R22
          brne   DEL302            ; ~50 ms
          dec    R23
          brne   DEL303            ; ~1 sec
          ldi    R21,255
DEL305:   rcall  delay1mks
          rcall  delay1mks
          rcall  delay1mks
          rcall  delay1mks
          dec    R21
          brne   DEL305            ; 1 sec

          in     R21,PortLED
          ldi    R22,(1<<LED)      ; LED: PORTB,5
          eor    R21,R22           ; invert LED state
          out    PortLED,R21
          wdr                      ; RESET WDT

          dec    R20
          brne   DEL304            ; 30 sec
          rjmp   RDDHT             ; New read from DHT22

; wait WDT up to 30 sec or jmp 0000
ERRORDHT: ldi    R20,30
DELER4:   ldi    R23,20
DELER3:   ldi    R22,200
DELER2:   ldi    R21,210
DELER1:   rcall  delay1mks
          dec    R21
          brne   DELER1            ; 250 mks
          dec    R22
          brne   DELER2            ; 50 ms
          dec    R23
          brne   DELER3            ; 1 sec
          dec    R20
          brne   DELER4            ; 30 sec
          rjmp   0x0000

; ================= Subroutines ===================
; UART send byte
; Input: R21 = byte
; Used: R22
TXUART:   lds    R22,UCSR0A
          sbrs   R22,UDRE0         ; Wait for empty transmit buffer
          rjmp   TXUART
          sts    UDR0,R21          ; Put data into buffer, sends the data
          ret

; Delay 18 ms
; Used: R21,R22
delay18ms: ldi    R22,72
del182:    ldi    R21,211
del181:    rcall  delay1mks
           dec    R21
           brne   del181           ; ~250 mks
           dec    R22
           brne   del182           ; ~18 ms
           ret

; Delay 1 mks
delay1mks: rjmp   PC+1
           rjmp   PC+1
           rjmp   PC+1
           rjmp   PC+1
           nop
           ret

